class QueueSettings {
  _queueCoolDownTime = 100;

  get QueueCoolDownTime() {
    return this._queueCoolDownTime;
  }

  set QueueCoolDownTime(value) {
    if (value < 50) {
      this._queueCoolDownTime = 50;
      return;
    }

    this._queueCoolDownTime = value;
  }

  constructor() {}

}

/**
 * This working like Task.Delay() on the C#
 * @param ms
 * @returns
 * @example `
 *  doSomething...
 *  wait(1000) // One second
 *  continue...
 *  `
 */
const wait = ms => {
  return new Promise(resolve => setTimeout(resolve, ms));
};

/**
 * # Simple Queue System
 * ### Methods
 * - PushMessage
 * - Comsume
 * - ACK
 * - Clear
 * @example
 * `Define your queue`
 * new SimpleQueue<MessageObject>();
 * @example
 * `Publish a message`
 * new SimpleQueue<MessageObject>().Push(MessageObject);
 * @example
 * `Full example`
 * var queue = new SimpleQueue<MessageObject>();
 * queue.Consume(async (message: MessageObject) => {
 *  do something...
 * });
 * queue.Push(MessageObject);
 */

class BasicQueue {
  _messages = [];
  Handler = null;
  AutoACK = false;
  _queueStarted = false;
  _currentMessage = null;
  Settings;

  constructor(settings = null) {
    if (settings === null) this.Settings = new QueueSettings();else this.Settings = settings;
  }

  Ack() {
    if (this._currentMessage === null) return;

    const i = this._messages.indexOf(this._currentMessage);

    if (i !== -1) this._messages.splice(i, 1);
    this._currentMessage = null;
    this._queueStarted = this._messages.length > 0;
  }

  PushMessage(message) {
    this._messages.push(message);

    this.start();
  }

  Consume(handler, autoAck) {
    this.Handler = handler;
    this.AutoACK = autoAck;
  }

  get MessageCount() {
    return this._messages.length;
  }

  Clear() {
    let removedMessages = this._messages.filter(x => !x.UnAck);

    this._messages = this._messages.filter(x => !removedMessages.includes(x));
    return removedMessages;
  }

  async start() {
    if (this._queueStarted) return;
    this._queueStarted = true;

    while (this._queueStarted) {
      this._currentMessage = this._messages[0];
      await wait(this.Settings.QueueCoolDownTime);
      if (this.Handler === null || this._currentMessage.UnAck === true) continue;

      try {
        this._currentMessage.UnAck = true;
        await this.Handler(this._currentMessage);
      } catch (error) {
        console.error("Consumer Error Ocurred. See the details; \n", error);
      }

      if (this.AutoACK) this.Ack();
    }
  }

}

export { BasicQueue as default };
